/*
    asm.h

    The API for the assembler and virtual machine is not thread-safe,
    but it does permit multiple VMs (each with its own assembled program)
    that can be called on demand. Don't call more than one simultaneously.

    These are the API function prototypes for the assembler and VM,
    given in approximate order of us.
*/

int alu_trace;                      // 0 nothing, 1 panel, 2 detail

long alu                            // Run the VM's ALU on one computation.
    (int opcode, long l, long r, int *flags);
void alu_setup(void);               // Prepare VM's ALU for use.


void *asm_assemble                  // Create a VM and assemble a program.
    (char *source, char *flags, char *banner);
void asm_disassemble                // Disassemble a word.
    (char *prefix, long code, char *suffix);
void asm_reset(void *context);      // Reset all regs so program can restart.
void asm_run(void *context);        // Run the assembled program.
long *asm_get_register              // Get pointer to a named register in VM.
    (void *context, char *name);
int *asm_get_flags(void *context);  // Get pointer to flags in VM.
int asm_fetch_code                  // Copy assembled code to external buffer.
    (void *context, long *out, int limit);
void asm_dealloc(void *context);    // Destroy and deallocate a VM.


