/*
    context.h

    The API for the assembler and virtual machine is not thread-safe,
    but it does permit multiple VMs (each with its own assembled program)
    that can be called on demand. Don't call more than one simultaneously.

    The function prototypes for calling the assembler are in asm.h, and
    employ an opaque "context" pointer for specifying which VM is in use.
    When these functions are entered, that pointer gets copied into
    a static struct context * named 'X', which everything happens via.

    Here is the definition for this 'X'.
*/

enum { max_call_depth = 16 };
static struct context {
    char *src;                      // source code
    int free_src;                   // dealloc source code with context?
    int list;                       // provide listing?

    long *code;                     // assembled code
    int n_code;                     // next free location in code segment

    long *regs;                     // registers
    int n_regs;                     // next free register (0-511)

    struct sym *sym_root;           // symbol table
    struct eMsg *eMsg_head;         // parse errors
    struct eMsg **eMsg_tail;        // initialize to &eMsg_head

    int ip;                         // instruction pointer
    int flags;                      // ALU flags

    int stack[max_call_depth];      // return address stack
    int n_stack;                    // depth of stack
} *X;
