#!/usr/bin/python3

# h = 1/2
# h    = .1000000000000000   16 bits
# h**h = .10110101            8 bits
# **2  = .0111111111111001   approximation

def perm():
    avail = list(range(64))
    done = []
    while avail:
        perm.fract *= len(avail)
        rand = perm.fract >> root_bits
        perm.fract &= root_mask
        done.append(avail[rand])
        del avail[rand]
    return tuple(done)

def boxes():
    return tuple(perm() for i in range(64))

def row():
    return tuple(boxes() for i in range(6))

def run():
    return row(), row(), row()

root_bits = 341000           # needed = 18 * 64 * (64!).bit_length()
root_mask = (1 << root_bits) - 1
half_bits = 2 * root_bits    # must be even
half = 1 << half_bits - 1    # 1/2 in this radix
root = 3 << root_bits - 2    # initial estimate for (1/2)**(1/2)

print('Computing %s bits of sqrt(2)/2 ' % root_bits, end='')
while abs(half - root * root) > 2 * root:
    root = (root + half // root) >> 1
    print(end = '.', flush = True)
print(' done!')
perm.fract = root

v = run()                    # v is the output of this run
# print(v[2][5][63])         # show last permutation as a check

# Extension to output to C compiler

def out_row(name, a):
    print('\nchar *%s_sbox[6][64] = {' % name, file = f)
    for i, row in enumerate(a):
        for r in row:
            # this encoding skips \ and "
            s = ''.join(map(lambda x: chr(32 + x + (x>=2) + (x>=59)), r))

            # diagnostic encoding uses \ and "
            # s = ''.join(map(lambda x: chr(32 + x), r))
            # s = s.replace('\\', '\\\\').replace('"', '\\"')

            print('\t"%s",' % s, file = f)
        if i != 5: print(file = f)
    print("};", file = f);

f = open('sbox.c', 'w')
print("/* automatically generated by gen-sbox.py */", file = f)
out_row('alpha', v[0])
out_row('beta',  v[1])
out_row('gamma', v[2])
f.close()
