/*
    This is a long story. Please document. TODO
    MWA 2021 Jan 31 - Feb 2
*/

#define ifnot(x) if (!(x))

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

enum {
    n_in = 4,           // number of input bits (limit: 5)
    too_long = 0xff,    // delays uninitialized
    too_many = 0xff,    // number of components uninitialized
    width =             // 2 ** n_in
        ((size_t) 1) << n_in,
};

enum types {
    gate_and = 'a', gate_or = 'o', gate_nand = 'A', gate_nor = 'O',
    gate_not = 'i', gate_xor = 'x', gate_identity = 'I',
    gate_free_or = 'f', gate_free_or_not = 'F',
};

size_t height =         // 2 ** width
    ((size_t) 1) << width;

#define mask ((((size_t) 1) << width) - 1)

size_t n_found;         // number of known functions

struct fn {
    uint32_t left;         // left ancestor of this function
    uint32_t right;        // right ancestor of this function
    uint8_t  delays;       // time to reach this function, units of .1 ns
                           // XXX there is a circuit limit of 25.5 ns
    uint8_t  n_pins;       // # pins needed to compute this function
    uint8_t  last_op;      // last operation to reach this function
} *fns, *free_ors, *free_or_nots;

uint32_t *found;            // list of calculated function indices

struct fn init_fn = { mask, mask, 255, 255, 255 };

int weigh_delay = 1;
int weigh_parts = 1;

int8_t gate_delay[128];     // indexed by enum types, units 0.1 ns
int8_t gate_pins[128];      // indexed by enum types, amortized to AUC2G pins

// Cost function is fixed for now.
int cost(int delay, int parts)
{
    return weigh_delay * delay + weigh_parts * parts;
}

// Do one gate.
void one_gate(
    int gate,           // 'a' = AND, etc.
    uint32_t fn,        // proposed function
    uint32_t l,         // proposed left ancestor
    uint32_t r,         // proposed right ancestor
    struct fn *tfn,     // table proposed function belongs in
    struct fn *tl       // table left ancestor belongs in
) {                     // right ancestor is always in 'fns' table
    struct fn *lp = tl + l, *rp = fns + r, *fp;
    int delay = lp -> delays, t;
    int parts = lp -> n_pins + rp -> n_pins;

    fn &= mask;         // always
    fp = tfn + fn;

    // time to compute proposed fn?
    t = rp -> delays;
    if (t > delay)
        delay = t;
    delay += gate_delay[gate];

    // number of parts to compute proposed fn?
    if (gate == gate_not)
        parts >>= 1;    // count only one inverter input
    if (gate != gate_free_or && gate != gate_free_or_not)
        parts += 1;     // certain gates effectively have no parts

    if (cost(delay, parts) >= cost(fp -> delays, fp -> n_pins))
        return;         // this proposal does not improve our cost

    // if no previous implementation of fn, note its discovery
    if (fp -> delays == too_long)
        found[n_found++] = fn;

    // this is a new cheapest implementation of fn
    fp -> left = l;
    fp -> right = r;
    fp -> last_op = gate;
    fp -> delays = delay;
    fp -> n_pins = parts;
}

// Use combinations of what we've found to find more.
void solve(void)
{
    size_t li, ri;
    uint32_t l, r;

    // This loop construct relies on gate inputs being symmetric.
    // This loop construct is too ambitious for n_in = 5.
    // We may look at that later.

    // Every function found will eventually be a left input.
    for (li = 0; li < n_found; li++) {
        fprintf(stderr, "\r%lu ", (unsigned long) li);
        l = found[li];
        one_gate(gate_not, ~l, l, l, fns, fns);              // NOT

        // Consider possible right inputs.
        // Every left/right *set* will eventually be processed.
        for (ri = 0; ri < li; ri++) {
            r = found[ri];
            one_gate(gate_and, l & r, l, r, fns, fns);       // AND
            one_gate(gate_or, l | r, l, r, fns, fns);        // OR
            one_gate(gate_nand, ~(l & r), l, r, fns, fns);   // NAND
            one_gate(gate_nor, ~(l | r), l, r, fns, fns);    // NOR
            one_gate(gate_xor, l ^ r, l, r, fns, fns);       // XOR
        }
    }

    fprintf(stderr, "\r");
}

// Synchronous RAMs have three enable lines named E1#, E2, E3#.
// Two are active-low, one is active-high. When driving E1#,
// these RAMs offer a free "OR" function at E3# and a free
// "OR NOT" on E2. Compute optimal-cost solutions for these.
void freebies(void)
{
    uint32_t l, r;

    // free OR function
    memcpy(free_ors, fns, height * sizeof(*fns));
    for (l=1; l<height; l++)
        for (r=0; r<l; r++)
            one_gate(gate_free_or, l | r, l, r, free_ors, fns);

    // free OR NOT function
    memcpy(free_or_nots, free_ors, height * sizeof(*fns));
    for (l=0; l<height; l++)
        for (r=0; r<height; r++) {
            if (r == l) continue;
            one_gate(gate_free_or_not, l | ~r, l, r, free_or_nots, free_ors);
        }
}

void show_how(uint32_t fn, int level, struct fn *hier)
{
    size_t i;
    struct fn *fp = hier + fn;
    struct fn *left_ancestor = fp -> last_op == 'F' ? free_ors : fns;

    // Terminal symbol?
    for (i=0; i < n_in; i++)
        if (fn == found[i]) {
            printf("%c", 'a' + (int) i);
            return;
        }

    if (level)
        printf("(");

    // Inverter?
    if (fp -> last_op == 'i') {
        printf("not ");
        show_how(fp -> left, level+1, fns);
        goto done;
    }

    // Operations on two inputs
    show_how(fp -> left, level+1, left_ancestor);
    switch (fp -> last_op) {
        case 'a': printf(" and "); break;
        case 'o': printf(" or "); break;
        case 'A': printf(" nand "); break;
        case 'O': printf(" nor "); break;
        case 'x': printf(" xor "); break;
        case 'f': printf(" free_or "); break;
        case 'F': printf(" free_or_not "); break;
        default: printf(" ERROR "); break;
    }
    show_how(fp -> right, level+1, fns);

done:
    if (level)
        printf(")");
}

// Show us what we know.
void show(int order_found, int how, struct fn *hier)
{
    size_t n, fn;

    for (n = 0; n < height; n++) {
        fn = order_found ? found[n] : n;
//      printf("%01lx  %.1f ns", (unsigned long) fn,
        printf("%02lx  %.1f ns", (unsigned long) fn,
//      printf("%04lx  %.1f ns", (unsigned long) fn,
            hier[fn].delays / 10.);
        if (how) {
            printf("  ");
            show_how(fn, 0, hier);
        }
        printf("\n");
    }
}

// Compact RPN representation of fn.
void unfurl(size_t fn, FILE *f, struct fn *hier)
{
    size_t i;
    struct fn *fp = hier + fn;
    struct fn *left_ancestor = fp -> last_op == 'F' ? free_ors : fns;

    // Terminal symbol? (Terminals are digits 0 through at most 4.)
    for (i=0; i < n_in; i++)
        if (fn == found[i]) {
            fprintf(f, "%i", (int) i);
            return;
        }

    // Inverter? (Suffix 'i'.)
    if (fp -> last_op == gate_not) {
        unfurl(fp -> left, f, fns);
        fprintf(f, "%c", fp -> last_op);
        return;
    }

    // Operations on two inputs (suffix to match operation).
    unfurl(fp -> left, f, left_ancestor);
    unfurl(fp -> right, f, fns);
    fprintf(f, "%c", fp -> last_op);
}

void save_compact(char *fname, struct fn *hier)
{
    FILE *f = fopen(fname, "w");
    size_t fn;

    if (!f) {
        fprintf(stderr, "unable to write %s\n", fname);
        return;
    }
    for (fn = 0; fn < height; fn++) {
        fprintf(f, "%i ", (int) hier[fn].delays);
        unfurl(fn, f, hier);
        fprintf(f, "\n");
    }

    fclose(f);
}

void do_over(void)
{
    uint32_t bit, i, fn;
    size_t loop; 
    struct fn *fp;

    // Initialize.
    n_found = 0;                // would be needed to reinitialize
    for (loop = 0; loop < height; ++loop)
        fns[loop] = init_fn;

    // Create identity functions.
    for (i=0; i<n_in; i++) {
        fn = 0;
        for (bit = 0; bit < width; ++bit)
            if (bit & (1 << i))
                fn |= 1 << bit;

        fp = fns + fn;
        fp -> left = fn;          // means nothing prior
        fp -> right = fn;         // means nothing prior
        fp -> last_op = gate_identity;
        fp -> delays = 0;         // no cost to compute
        fp -> n_pins = 0;         // no cost to compute
        found[n_found++] = fn;    // available for use!
    }

    // Iteratively find the other functions.
    solve();
}

int main(int argc, char *argv[])
{
    // establish gate delays and pins per gate
#if 1       // SN74AUC2Gxx into 15 pF, units of 0.1 ns
    gate_delay[gate_and] = 10;
    gate_delay[gate_or] = 11;
    gate_delay[gate_nand] = 10;
    gate_delay[gate_nor] = 13;
    gate_delay[gate_not] = 12;
    gate_delay[gate_xor] = 13;
    gate_pins[gate_and] = 4;
    gate_pins[gate_or] = 4;
    gate_pins[gate_nand] = 4;
    gate_pins[gate_nor] = 4;
    gate_pins[gate_not] = 3;
    gate_pins[gate_xor] = 4;
#else       // SN74AUC1Gxx into 15 pF, units of 0.1 ns
    gate_delay[gate_and] = 16;
    gate_delay[gate_or] = 14;
    gate_delay[gate_nand] = 14;
    gate_delay[gate_nor] = 17;
    gate_delay[gate_not] = 16;
    gate_delay[gate_xor] = 13;
    gate_pins[gate_and] = 5;
    gate_pins[gate_or] = 5;
    gate_pins[gate_nand] = 5;
    gate_pins[gate_nor] = 5;
    gate_pins[gate_not] = 4;
    gate_pins[gate_xor] = 5;
#endif

    // Get memory.
    fns = malloc(height * sizeof(*fns));
    free_ors = malloc(height * sizeof(*free_ors));
    free_or_nots = malloc(height * sizeof(*free_or_nots));
    found = malloc(height * sizeof(*found));

    ifnot (fns && free_ors && free_or_nots && found) {
        fprintf(stderr, "out of memory\n");
        return -1;
    }

    weigh_delay = 10, weigh_parts = 1;
    do_over();
    save_compact("speed", fns);
    freebies();
    save_compact("speed-freebies", free_or_nots);


    weigh_delay = 1, weigh_parts = 10;
    do_over();
    save_compact("size", fns);
    freebies();
    save_compact("size-freebies", free_or_nots);

    // Show us the last set of results.
    show(0, 1, free_or_nots);

    // We have what we need. We skip undoing the mallocs.
    return 0;
}
