#!/usr/bin/python3

# MWA Why do we not use any inverters ever?

from array import array
from itertools import repeat

max_input_bits = 4              # input bits allowed in an expression
width = 1 << max_input_bits     # width of an expression truth table
height = 1 << width             # number of functions possible

rpnDB = {}
costDB = {}

AND = '''
; Comments like this
AND gate (example)
; title is REQUIRED
B | A || AB
0 0  0
0 1  0
1 0  0
1 1  1
'''

adder = '''
Full Adder (example)
Ci | B | A || Co | S
0 0 0  0 0
0 0 1  0 1
0 1 0  0 1
0 1 1  1 0
1 0 0  0 1
1 0 1  1 0
1 1 0  1 0
1 1 1  1 1
'''

alpha_lockout = '''
Alpha lockout
d111 | d112 | d121 | i100 || e100 | e101 | e113 | e115 | e120 | e121
0 0 1 1    0 0 1 1 1 1
1 1 1 1    1 0 1 1 0 0
1 0 1 1    1 0 0 0 1 0
X X 0 1    0 0 0 1 1 1
X X X 0    0 0 1 1 0 0
0 1 1 1    X 1 1 1 1 1
'''

alpha_lockout_solution_feb_3 = '''
; 13 gates
e100 (2.1 ns) = d111 d112 or d121 i100 and *and*
e101 (3.3 ns) = d121i100and d112 nand d111 *nor*
e113 (2.0 ns) = d112d121and d111d121nand d121 and free-or i100 free-or-not
e115 (3.1 ns) = d121i100and d111 nand d112 or
e120 (3.0 ns) = d112 d121 and d111 nand i100 *and*
e121 (2.0 ns) = d111 d121 nand i100 and

*Double these gates due to high fanout after longer computation.
'''

beta_lockout = '''
Beta lockout
d113 | d114 | i101 | Mwp || e102 | e103 | e111 | e110 | e118 | e129
X X 0 X    0 0 1 1 1 0
0 0 1 X    0 0 1 1 1 1
0 1 1 1    0 0 1 1 1 1
0 1 1 0    0 1 0 0 1 1
1 0 1 X    1 0 0 1 1 1
1 1 1 X    1 0 1 1 0 1
'''

beta_lockout_solution_feb_4b = '''
; 11 gates
e102 (1.0 ns) = d113 i101 and
e103 (2.3 ns) = d113 Mwp nor d114 i101 and *and*
e111 (3.0 ns) = d114 Mwp nand i101 and d113 d114 xor nand
e110 (2.2 ns) = d113 Mwp or d114 i101 nand or
e118 (2.0 ns) = d114i101and d113 *nand*
e129 (0.0 ns) = i101
'''

gamma_lockout = '''
Gamma lockout
d117 | d118 | i103 || e106 | e108 | e107 | e109 | e128 | e127
0 0 1    0 0 0 0 1 1
0 1 1    0 1 X 0 1 1
1 0 1    1 0 1 0 1 0
1 1 1    1 X 1 1 1 0
X X 0    0 1 1 0 0 0
'''

gamma_lockout_solution_may_19 = '''
; 8 gates
e106 (1.0 ns) = d117 i103 and
e108 (0.0 ns) = d118 i103 free-or-not
e107 (2.3 ns) = d117 d118 nor i103 nand
e109 (2.0 ns) = d117 i103 and d118 and
e128 (0.0 ns) = i103
e127 (2.0 ns) = d117 i103 nand i103 and
'''

ip_lockout = '''
IP lockout
d119 | d120 | branch || e114 | e122 | e117 | e119
0 0 X    0 1 1 1
0 1 X    1 0 1 1
1 X 0    1 1 0 1
1 X 1    1 1 1 0
'''

ip_lockout_solution_feb_3 = '''
; 5 gates
e114 (1.1 ns) = d119 d120 or
e122 (2.0 ns) = d119 d120 nand d120 nand
e117 (2.0 ns) = d119 branch nand d119 nand
e119 (1.0 ns) = d119branchnand
'''

c_lockout = '''
C lockout
d115 | d116 || e105 | e104 | e123 | e124 | e116
0 0    0 1 1 1 1
0 1    0 0 0 0 1
1 X    1 1 0 1 0
'''

c_lockout_solution_feb_3 = '''
; 4 gates
e105 (0.0 ns) = d115
e104 (2.0 ns) = d115 d116 nand d116 nand
e123 (1.3 ns) = d115 d116 nor
e124 (2.0 ns) = d115d116nandd116nand
e116 (1.2 ns) = d115 not
'''

z_lockout = '''
Z lockout
d133 | i105 || e130
0 1   0
1 1   1
X 0   0
'''

z_lockout_solution_feb_6 = '''
; 1 gate
e130 (1.0 ns) = d133 i105 and
'''

s_lockout = '''
S lockout
d134 | d135 || e131
1 0   0
1 1   1
0 X   0
'''

s_lockout_solution_feb_6 = '''
; 1 gate
e131 (1.0 ns) = d134 d135 and
'''

e_lockout = '''
E lockout
d136 | d137 || e125 | =d136 | e126 | e132
1 0  0 1 1 1
1 1  1 1 1 1
0 0  0 0 0 1
0 1  0 0 1 0
'''

e_lockout_solution_feb_6 = '''
; 4 gates
e125 (1.0 ns) = d136 d137 and
e126 (1.1 ns) = d136 d137 or
e132 (2.0 ns) = d136 d137 nand d137 *nand*
'''

d_lockout = '''
D lockout
i106 | i102 || =i106 | =i102 | e133
0 1	 0 1 1
0 0	 0 0 0
1 X	 1 X 0
'''

d_lockout_solution_june_9 = '''
; 1 gate
=i106 (0.0 ns) = i106
=i102 (0.0 ns) = i102
e133 (2.0 ns) = i106 i102 nand i102 and
'''

def loadSolutions():
    for fName in rpnDB:
        rpn = rpnDB[fName]
        costs = costDB[fName]
        del rpn[:]
        del costs[:]
        for line in open(fName).readlines():
            c, r = line.strip().split(' ')
            costs.append(int(c))
            rpn.append(r)
        if len(costs) != height:
            print('expected %i lines in %s' % (height, fName))
            raise SystemExit

def ttParse(tt):
    outVars = []
    fns = []
    lines = [l.split(';')[0].strip() for l in tt.splitlines()]
    lines = [l for l in lines if l]

    title = lines.pop(0)
    head = lines.pop(0)
    # head now looks like 'B | A || AB'

    try:
        inVars, outVars = head.split('||')
    except:
        print('Title or column heading problem: %s' % title)
        raise SystemExit

    inVars = [v.strip() for v in inVars.split('|')]
    outVars = [v.strip() for v in outVars.split('|')]
    if len(inVars) > max_input_bits:
        print('%i is too many input bits' % len(inVars))
        raise SystemExit
    # inVars now looks like ['B', 'A']
    # outVars now looks like ['AB']

    # output truth tables are represented as ints
    # coverage of input values is a count of hits
    # truth table must cover all input combinations exactly once
    tts = [0] * len(outVars)
    coverage = [0] * width

    for ll in lines:
        # ll now looks like '   X 0 1 1  1 0  '
        l = ll.lower().split()
        if len(l) != len(inVars) + len(outVars):
            print('wrong number of fields in %s' % ll)
            raise SystemExit
        inbits = l[:len(inVars)]
        # if fewer than four inputs, fill to four
        inbits = (['x', 'x', 'x', 'x'] + inbits)[-4:]
        # change "don't care" output bits to ones (XXX could search instead)
        for i in range(len(inVars), len(l)):
            if l[i] == 'x':
                l[i] = '1'
        outbits = l[len(inVars):]
        if set(inbits) - {'0', '1', 'x'}:
            print('invalid input bits in %s' % ll)
        if set(outbits) - {'0', '1'}:
            print('invalid output bits in %s' % ll)
        outbits = list(map(int, outbits))
        # inbits now looks like ['X', '0', '1', '1']
        # outbits now looks like [1, 0]

        # go through all truth table values
        for i in range(width):
            if not matches(i, inbits):
                continue
            # this row applies if we get here
            coverage[i] += 1
            for column, ob in enumerate(outbits):
                if ob:
                    tts[column] |= 1 << i

    # ensure every input case was covered exactly once
    if set(coverage) != {1}:
        print('over/undercovered truth cases %s' % coverage)
        raise SystemExit

    return title, inVars[::-1], outVars, tts

# i is an integer, bits looks like ['0', '1', 'x', 'x', 0 ]
# Bits of i are checked to see if they match 0, 1, or don't care.
# The array's order is msb to lsb. Doesn't matter, as long as
# this Python program keeps it consistent. The C side of the
# tool doesn't know or care.
def matches(i, bits):
    # without [::-1], would be lsb to msb
    for pos, bstr in enumerate(bits[::-1]):
        v = bool(i & 1 << pos)
        if bstr == '1' and not v:
            return False
        if bstr == '0' and v:
            return False
    return True

def solve(tTable):
    title, ivars, ovars, truths = ttParse(tTable)

    d = { 'a' : 'and', 'o' : 'or', 'i': 'not', 'x' : 'xor',
        'A' : 'nand', 'O' : 'nor', 'f' : 'free-or', 'F' : 'free-or-not' }
    for i, iv in enumerate(ivars):
        d[str(i)] = iv

    print(title)
    print('-' * len(title))
    for ovar, truth in zip(ovars, truths):
        have = []
        for variant in rpnDB:
            rpn = rpnDB[variant]
            costs = costDB[variant]
            fn = rpn[truth]
            ns = costs[truth] / 10
            text = ' '.join([d[c] for c in fn])
            if fn in have:
                continue
            have.append(fn)
            print('%s (%.1f ns) = %s' % (ovar, ns, text))
        print()

def run():
    for fName in 'size', 'speed', 'size-freebies', 'speed-freebies':
        rpnDB[fName] = []           # RPN solutions by truth table
        costDB[fName] = array('I')  # solution cost by truth table

    loadSolutions()
    for table in [AND, adder, alpha_lockout, beta_lockout, gamma_lockout,
            ip_lockout, c_lockout, z_lockout, s_lockout, e_lockout,
            d_lockout]:
        solve(table)

run()
