/*
    c_244.c

    This is the simulation's implementation of a Texas Instruments
    SN74AUC16244 16-bit buffer. It's architected with four 4-bit sections.
*/

// Here is a four-bit slice of the part.
typedef struct {
    in_pin oe_;             // active-low output enable
    in_pin a[4];            // input pins
    out_pin y[4];           // output pins
} c_244_quarter;

// Here is the entire part.
typedef struct {
    Prefix pre;             // fields common to all components
    c_244_quarter buf[4];   // four 4-bit slices
} c_244;

// Here are names and byte offsets of input pins within the struct.
in_namer in_namer_244[] = {
    "1oe#", &((c_244 *) NULL)->buf[0].oe_,
    "a0", &((c_244 *) NULL)->buf[0].a[0],
    "a1", &((c_244 *) NULL)->buf[0].a[1],
    "a2", &((c_244 *) NULL)->buf[0].a[2],
    "a3", &((c_244 *) NULL)->buf[0].a[3],
    "2oe#", &((c_244 *) NULL)->buf[1].oe_,
    "a4", &((c_244 *) NULL)->buf[1].a[0],
    "a5", &((c_244 *) NULL)->buf[1].a[1],
    "a6", &((c_244 *) NULL)->buf[1].a[2],
    "a7", &((c_244 *) NULL)->buf[1].a[3],
    "3oe#", &((c_244 *) NULL)->buf[2].oe_,
    "a8", &((c_244 *) NULL)->buf[2].a[0],
    "a9", &((c_244 *) NULL)->buf[2].a[1],
    "a10", &((c_244 *) NULL)->buf[2].a[2],
    "a11", &((c_244 *) NULL)->buf[2].a[3],
    "4oe#", &((c_244 *) NULL)->buf[3].oe_,
    "a12", &((c_244 *) NULL)->buf[3].a[0],
    "a13", &((c_244 *) NULL)->buf[3].a[1],
    "a14", &((c_244 *) NULL)->buf[3].a[2],
    "a15", &((c_244 *) NULL)->buf[3].a[3],
    NULL, NULL,
};

// Here are names and byte offsets of output pins within the struct.
out_namer out_namer_244[] = {
    "y0", &((c_244 *) NULL)->buf[0].y[0],
    "y1", &((c_244 *) NULL)->buf[0].y[1],
    "y2", &((c_244 *) NULL)->buf[0].y[2],
    "y3", &((c_244 *) NULL)->buf[0].y[3],
    "y4", &((c_244 *) NULL)->buf[1].y[0],
    "y5", &((c_244 *) NULL)->buf[1].y[1],
    "y6", &((c_244 *) NULL)->buf[1].y[2],
    "y7", &((c_244 *) NULL)->buf[1].y[3],
    "y8", &((c_244 *) NULL)->buf[2].y[0],
    "y9", &((c_244 *) NULL)->buf[2].y[1],
    "y10", &((c_244 *) NULL)->buf[2].y[2],
    "y11", &((c_244 *) NULL)->buf[2].y[3],
    "y12", &((c_244 *) NULL)->buf[3].y[0],
    "y13", &((c_244 *) NULL)->buf[3].y[1],
    "y14", &((c_244 *) NULL)->buf[3].y[2],
    "y15", &((c_244 *) NULL)->buf[3].y[3],
    NULL, NULL,
};

// Here are maximum times from the datasheet for the SN74AUC16244.
enum {
    t_pd = 1800,
    t_en = 1900,
    t_dis = 2000,
};

// We do nothing to preset a 16244 at program start.
void init_16244(UNUSED void *v, UNUSED int z)
{
}

// Simulate one of the four-bit slices.
void be_16244_quarter(c_244_quarter *q, int trace, int unit)
{
    int i;

    // They have separate output enable pins:
    if (q->oe_.changed) {
        q->oe_.changed = 0;

        // Four bits to enable or disable
        if (q->oe_.val == lo)
            output_impedance_after_delay(q->y, -4, t_en);
        else
            output_impedance_after_delay(q->y, 4, t_dis);
    }

    // And here are the ordinary pins:
    for (i = 0; i < 4; i++) {
        if (!q->a[i].changed)
            continue;
        q->a[i].changed = 0;
        pin_output_after_delay(&q->y[i], q->a[i].val, t_pd);
        if (trace)
            printf("%s going %s\n", q->y[i].name, pin_val_text(q->a[i].val));
    }
}

// Simulate the entire 16-bit buffer chip.
void be_16244(void *v)
{
    c_244 *b = v;
    int i;

    // There are four units of four bits each.
    for (i = 0; i < 4; i++)
        be_16244_quarter(&b->buf[i], b->pre.trace, i);
}
