/*
    c_pull.c

    This is the simulation's thoroughly broken implementation of a
    pull resistor. Specifically:

    a) The resistor is polarized. 'a' is in, 'y' is out.
       I don't consider this to actually be broken, but it's
       unusual enough to write up here.

    b) It's a single resistor. The Dauug|36 architecture uses pull
       resistor arrays instead. This is most broken.

    c) The simulation does not actually emulate resistors at all.

    d) The simulation does not support any weak logic levels that
       can be "pulled away from".

    e) The generated netlist creates pull resistor arrays (badly
       having the same name as this single-resistor implementation),
       but does not make any connections to these components.

    f) The resistance value is considered to be zero ohms, so it
       essentially works as a jumper.

    g) There is actually a regression test script for this that
       yields a passing score, based on the zero ohm jumper model.
*/

typedef struct {
    Prefix pre;             // fields common to all components
    in_pin a;               // connect to a power rail
    out_pin y;              // output
} c_pull;

// Here is the name and byte offset of the struct's one input pin.
in_namer in_namer_pull[] = {
    "a", &((c_pull *) NULL)->a,
    NULL, NULL,
};

// Here is the name and byte offset of the struct's one output pin.
out_namer out_namer_pull[] = {
    "y", &((c_pull *) NULL)->y,
    NULL, NULL,
};

// No initialization is needed.
void init_pull(void *v, int truth)
{
}

// There are no adjustments supported during simulation.
void tweak_pull(void *v, char *parm)
{
}

// Here is the simulation.
void be_pull(void *v)
{
    c_pull *g = v;

    if (!g->a.changed)
        return;
    g->a.changed = 0;

    // The output logic value matches the input.
    pin_output_after_delay(&g->y, g->a.val, 1);

    if (g->pre.trace)
        printf("%15li ps %s.y going %s\n",
                now, g->pre.name, pin_val_text(g->a.val));
}
