/*
    c_switch.c

    This is the simulation's implementation of a mythical DIP switch
    that outputs low, high, uncertain (transition pending), and
    high-impedance. The default output is high-impedance. This is a
    Very Good Switch, because its delay is just 1 picosecond. It may
    be suitable as an add-on component for certain test runs.
*/    

// single DIP switch (with output enable; in fact the default is hi_z)
typedef struct {
    Prefix pre;             // fields common to all components
    int touch;              // -1 (no change), 'lo', 'hi', 'uncert', 'hiz'
    out_pin y;              // output from switch
} c_switch;

// This component has no input pins.
in_namer in_namer_switch[] = {
    NULL, NULL,
};

// Here is the name and byte offset of the struct's one output pin.
out_namer out_namer_switch[] = {
    "y", &((c_switch *) NULL)->y,
    NULL, NULL,
};

// We start this switch up in high impedance.
void init_switch(void *v, int truth)
{
    c_switch *g = v;

    g->touch = hiz;
}

// For this tweak, 'parm' can be "hi", "lo", "hi_z", or "uncert".
void tweak_switch(void *v, char *parm)
{
    c_switch *g = v;
    int val = pin_text_val(parm);

    switch (val) {
        case lo:
        case hi:
        case uncert:
        case hiz:
            g->touch = val;
            break;

        default:
            ng("!bad switch tweak %s", parm);
            return;
    }
}

// Here we simulate a mythical switch.
void be_switch(void *v)
{
    c_switch *g = v;

    if (g->touch < 0)
        return;

    output_impedance_after_delay
        (&g->y, g->touch == hiz ? 1 : -1, 1);

    if (g->touch != hiz)
        pin_output_after_delay(&g->y, g->touch, 1);

    g->touch = -1;
}
